//+------------------------------------------------------------------+
//|                                       MeanReversionUtilities.mqh |
//|                                  Copyright 2024, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2024, MetaQuotes Ltd."
#property link      "https://www.mql5.com"
#include<Math\Stat\Math.mqh>
#include<VectorMatrixTools.mqh>
#include<OLS.mqh>

//+------------------------------------------------------------------+
//| Enumeration                                                      |
//+------------------------------------------------------------------+
enum ENUM_SPREAD_METHOD
{
 SPREAD_REG,//spread calculated as regression residuals
 SPREAD_DIST//spread calculated using distance formula a - xb
};
//+------------------------------------------------------------------+
//|Calculate Half life of Mean reversion                             |
//+------------------------------------------------------------------+
double mean_reversion_half_life(vector &data , double &lambda)
{
//---
 vector yy,zz;
 matrix xx;
//--- 
 OLS ols_reg;
//---  
   yy.Init(data.Size()-1,slice,data,1,data.Size()-1,1);
//--- 
   zz.Init(data.Size()-1,slice,data,0,data.Size()-2,1);
//---
   if(!xx.Init(zz.Size(),2) || !xx.Col(zz,0) || !xx.Col(vector::Ones(zz.Size()),1) || !ols_reg.Fit(yy-zz,xx))
    {
     Print(__FUNCTION__," Error in calculating half life of mean reversion ", GetLastError());
     return 0;
    }
//---  
   vector params = ols_reg.ModelParameters();
   lambda = params[0];
//---
   return (-log(2)/lambda);
//--- 
}
//+------------------------------------------------------------------+
//|Calculate Half life of Mean reversion                             |
//+------------------------------------------------------------------+
double mean_reversion_half_life(double &data[] , double &lbda)
{
//---
 vector in(data.Size(),assign,data);
//---
 if(in.Size()!=ulong(data.Size()))
  {
   Print(__FUNCTION__, " Critical error ");
   return 0;
  }
//---
 double half_life = mean_reversion_half_life(in,lbda);
//--- 
 return half_life;
//--- 
}
//+------------------------------------------------------------------+
//|Calculate the spread using distance forumla                       |
//+------------------------------------------------------------------+
bool get_spread_dist(double hedge_ratio,double &one[],double&two[],double&result[])
{
//---
 if(one.Size()!=two.Size())
   {
    Print("Input arrays not of same size");
    return false;
   }
//---
 if(ArrayResize(result,ArraySize(two))!=ArraySize(one))
   {
    Print(__FUNCTION__," error in computing the spread series. ");
    return false;
   }
//---  
   for(uint i =0; i<result.Size();i++)
      result[i]=one[i]-(hedge_ratio*two[i]);
//---
   return true;
}
//+------------------------------------------------------------------+
//|Calculate the spread  using regression                            |
//+------------------------------------------------------------------+
bool get_spread_reg(double &one[],double&two[],double&result[])
{
//---
 if(one.Size()!=two.Size())
   {
    Print("Input arrays not of same size");
    return false;
   }
//---
 if(ArrayResize(result,ArraySize(two))!=ArraySize(one))
   {
    Print(__FUNCTION__," error in computing the spread series. ");
    return false;
   }
//---
  vector y,rds;
  matrix matx;
//---  
   OLS ols;
//---   
   y.Init(10,assign,one,0,one.Size(),1);
//---
   matx.Init(10,2,olsdmatrix,two,0,two.Size(),1);
//---
   if( !y.Size() || matx.Rows() != y.Size() || !ols.Fit(y,matx))
     { 
       Print(__FUNCTION__," Critical error in calculation");
       return false;
     } 
//---
   rds = ols.Residuals(); 
   for(uint i = 0; i<result.Size(); i++)
        result[i]=rds[i]; 
//---
   return true;
}

//+------------------------------------------------------------------+
//|Hedge ratio calculation using standard deviation                  |
//+------------------------------------------------------------------+
double hedge_ratio_std(double &_one[],double&_two[], bool use_returns=false)
{
//---
 if(_two.Size()!=_one.Size() || (_one.Size()+_two.Size())<1)
   {
    Print(__FUNCTION__," Invalid function inputs ");
    return 0;
   }
//---
  if(use_returns)
   {
    double rets1[],rets2[];
    
    if(!MathDifference(_one,1,rets1) || !MathDifference(_two,1,rets2))
     {
      Print(__FUNCTION__," Differencing error ", GetLastError());
      return 0;
     }
    
    return MathStandardDeviation(rets1)/MathStandardDeviation(rets2); 
   }   
 else
   return MathStandardDeviation(_one)/MathStandardDeviation(_two); 
//--- 
}
//+------------------------------------------------------------------+
//| Hedge ratio calculation using regression                         |
//+------------------------------------------------------------------+
double hedge_ratio_reg(double &_one[],double&_two[])
{
//---
 if(_two.Size()!=_one.Size() || (_one.Size()+_two.Size())<1)
   {
    Print(__FUNCTION__," Invalid function inputs ");
    return 0;
   }
//---
  vector y,pms;
  matrix x;
//---  
   OLS ols;
//---   
   y.Init(10,assign,_one,0,_one.Size(),1);
//---
   x.Init(10,2,olsdmatrix,_two,0,_two.Size(),1);
//---
   if( !y.Size() || x.Rows() != y.Size() || !ols.Fit(y,x))
     { 
       Print(__FUNCTION__," Critical error in calculation");
       return 0;
     } 
//---
   pms = ols.ModelParameters();  
//---
   return pms[0];
//---  
}
//+------------------------------------------------------------------+
//|Calculates the reciprocal of a series in an array in place        |
//+------------------------------------------------------------------+
bool reciprocal(double &in_out[])
{
//---
 if(in_out.Size()<1)
  {
   Print(__FUNCTION__,"Array is empty");
   return false;
  }
//---
 for(uint i = 0; i<in_out.Size(); i++)
      in_out[i]=(in_out[i]!=0.0)?1.0/in_out[i]:in_out[i];
//---
 return true;      
}
//+------------------------------------------------------------------+
//| Check forex symbol pairs                                         |
//+------------------------------------------------------------------+
void checkFXsymbols(const int symbolprefixlength, const string quote,const string firstsymbol,const string secondsymbol, double &firstsymbolprices[], double &secondsymbolprices[])
{
//---
 int findfirst = StringFind(firstsymbol,quote);
//--- 
 int findsecond = StringFind(secondsymbol,quote);
//---
 if(findfirst < 0 || findsecond < 0)
   {
    Print("Warning: No common currency basis");
    return;
   }
//---
 findfirst-=symbolprefixlength;
 findsecond-=symbolprefixlength;
//---
 if(findfirst==0)
    reciprocal(firstsymbolprices);
//---
 if(findsecond==0)
    reciprocal(secondsymbolprices);   
//---
}